/*************** RATE-SCALING functions ***********************/


static u8 ipw_lower_rate_g[IPW_MAX_RATES] = {
	RATE_SCALE_5_5M_INDEX, 
	RATE_SCALE_5_5M_INDEX,
	RATE_SCALE_11M_INDEX, 
	RATE_SCALE_12M_INDEX,
	RATE_SCALE_18M_INDEX, 
	RATE_SCALE_24M_INDEX,
	RATE_SCALE_36M_INDEX, 
	RATE_SCALE_48M_INDEX, 
	IPW_INVALID_RATE,
	RATE_SCALE_1M_INDEX, 
	RATE_SCALE_2M_INDEX,
	RATE_SCALE_5_5M_INDEX,

};

static u8 ipw_higher_rate_g[IPW_MAX_RATES] = {
	RATE_SCALE_11M_INDEX, 
	RATE_SCALE_11M_INDEX,
	RATE_SCALE_18M_INDEX,
	RATE_SCALE_24M_INDEX, 
	RATE_SCALE_36M_INDEX,
	RATE_SCALE_48M_INDEX,
	RATE_SCALE_54M_INDEX, 
	IPW_INVALID_RATE,
	RATE_SCALE_2M_INDEX,
	RATE_SCALE_5_5M_INDEX, 
	RATE_SCALE_11M_INDEX,
	RATE_SCALE_12M_INDEX
};


static u8 ipw_lower_rate_g_prot[IPW_MAX_RATES] = {
	RATE_SCALE_5_5M_INDEX, 
	RATE_SCALE_5_5M_INDEX,
	RATE_SCALE_11M_INDEX,
	RATE_SCALE_11M_INDEX, 
	RATE_SCALE_18M_INDEX,
	RATE_SCALE_24M_INDEX,
	RATE_SCALE_36M_INDEX, 
	RATE_SCALE_48M_INDEX, 
	IPW_INVALID_RATE,
	RATE_SCALE_1M_INDEX, 
	RATE_SCALE_2M_INDEX,
	RATE_SCALE_5_5M_INDEX,
};

static u8 ipw_higher_rate_g_prot[IPW_MAX_RATES] = {
	RATE_SCALE_11M_INDEX, 
	RATE_SCALE_11M_INDEX,
	RATE_SCALE_18M_INDEX,
	RATE_SCALE_24M_INDEX, 
	RATE_SCALE_36M_INDEX,
	RATE_SCALE_48M_INDEX,
	RATE_SCALE_54M_INDEX, 
	IPW_INVALID_RATE,
	RATE_SCALE_2M_INDEX,
	RATE_SCALE_5_5M_INDEX, 
	RATE_SCALE_11M_INDEX,
	RATE_SCALE_18M_INDEX
};


static s32 ipw_expected_tpt_g[IPW_MAX_RATES] = {
	0, 0, 76, 104, 130, 168, 191, 202, 7, 13, 35, 58
};

static s32 ipw_expected_tpt_g_prot[IPW_MAX_RATES] = {
	0, 0, 0, 80, 93, 113, 123, 125, 7, 13, 35, 58
};

static s32 ipw_expected_tpt_a[IPW_MAX_RATES] = {
	40, 57, 72, 98, 121, 154, 177, 186, 0, 0, 0, 0
};

static s32 ipw_expected_tpt_b[IPW_MAX_RATES] = {
	0, 0, 0, 0, 0, 0, 0, 0, 7, 13, 35, 58
};

static struct ipw_tpt_entry ipw_tpt_table_a[] = {
	{-60, 22000, 0, 0, RATE_SCALE_54M_INDEX},
	{-64, 20000, 0, 0, RATE_SCALE_48M_INDEX},
	{-72, 18000, 0, 0, RATE_SCALE_36M_INDEX},
	{-80, 16000, 0, 0, RATE_SCALE_24M_INDEX},
	{-84, 12000, 0, 0, RATE_SCALE_18M_INDEX},
	{-85, 8000, 0, 0, RATE_SCALE_12M_INDEX},
	{-87, 7000, 0, 0, RATE_SCALE_9M_INDEX},
	{-89, 5000, 0, 0, RATE_SCALE_6M_INDEX}
};

static struct ipw_tpt_entry ipw_tpt_table_b[] = {
	{-86, 6000, 0, 0, RATE_SCALE_11M_INDEX},
	{-88, 3000, 0, 0, RATE_SCALE_5_5M_INDEX},
	{-90, 1000, 0, 0, RATE_SCALE_2M_INDEX},
	{-92, 800, 0, 0, RATE_SCALE_1M_INDEX}

};

static struct ipw_tpt_entry ipw_tpt_table_g[] = {
	{-60, 22000, 12000, 14000, RATE_SCALE_54M_INDEX},
	{-64, 20000, 11000, 13000, RATE_SCALE_48M_INDEX},
	{-68, 18000, 10000, 14000, RATE_SCALE_36M_INDEX},
	{-80, 16000, 9000, 11000, RATE_SCALE_24M_INDEX},
	{-84, 12000, 7000, 10000, RATE_SCALE_18M_INDEX},
	{-85, 8000, 5000, 8000, RATE_SCALE_12M_INDEX},
	{-86, 6000, 6000, 6000, RATE_SCALE_11M_INDEX},
	{-88, 3000, 3000, 3000, RATE_SCALE_5_5M_INDEX},
	{-90, 1000, 1000, 1000, RATE_SCALE_2M_INDEX},
	{-92, 800, 800, 800, RATE_SCALE_1M_INDEX}
};

static struct ipw_tpt_entry *ipw_get_tpt_by_rssi(s32 rssi, u8 mode)
{
	u32 index = 0;
	u32 table_size = 0;
	struct ipw_tpt_entry *tpt_table = NULL;

	if ((rssi < IPW_MIN_RSSI_VAL) || (rssi > IPW_MAX_RSSI_VAL))
		rssi = IPW_MIN_RSSI_VAL;

	switch (mode) {
	case MODE_IEEE80211G:
		tpt_table = ipw_tpt_table_g;
		table_size = ARRAY_SIZE(ipw_tpt_table_g);
		break;

	case MODE_IEEE80211B:
		tpt_table = ipw_tpt_table_b;
		table_size = ARRAY_SIZE(ipw_tpt_table_b);
		break;

	case MODE_IEEE80211A:
		tpt_table = ipw_tpt_table_a;
		table_size = ARRAY_SIZE(ipw_tpt_table_a);
		break;

	default:
		BUG();
		return NULL;
	}

	while ((index < table_size)
	       && (rssi < tpt_table[index].min_rssi))
		index++;

	index = min(index, (table_size - 1));

	return &tpt_table[index];
}


static void ipw_clear_window(struct ipw_rate_scale_data *window)
{
	window->data = 0;
	window->success_counter = 0;
	window->success_ratio = IPW_INVALID_VALUE;
	window->counter = 0;
	window->average_tpt = IPW_INVALID_VALUE;
	window->stamp = 0;
}

static int ipw_rate_scale_flush_old(struct ipw_priv *priv,
				    struct ipw_rate_scale_data *ctl)
{
	int rc = 0;
	s32 rate_index = 0;
	unsigned long flags;
	struct ipw_rate_scale_data *window;
	
	for (rate_index = 0; rate_index < IPW_MAX_RATES; rate_index++) {
		window = &(ctl[rate_index]);

		spin_lock_irqsave(&priv->lq_mngr.lock, flags);
		if (window->counter > 0) {
			if (time_after(jiffies, window->stamp + 
				       IPW_RATE_SCALE_WIN_FLUSH)) {
				IPW_DEBUG_RATE
				    ("flushed samples of rate index %d\n",
				     rate_index);
				ipw_clear_window(window);
			} else
				rc = 1;

		}
		spin_unlock_irqrestore(&priv->lq_mngr.lock, flags);
	}

	return rc;
}

#define IPW_RATE_SCALE_AVG_PACKETS      300
#define IPW_RATE_FLUSH_MAX              5000	// msec
#define IPW_RATE_FLUSH_MIN              50	// msec

static void ipw_bg_rate_scale_flush(unsigned long data)
{
	struct ipw_priv *priv = (struct ipw_priv *)data;
	int rc = 0, i;
	unsigned long flags;
	u32 tx_ave, duration;
	struct ipw_rate_scale_data *ctl;
	struct ipw_rate_scale_priv *scale_priv;
	struct sta_info *sta;
	struct ieee80211_local *local = wdev_priv(priv->net_dev->ieee80211_ptr);

	IPW_DEBUG_RATE("enter\n");

	if (priv->status & STATUS_EXIT_PENDING) {
		IPW_DEBUG_RATE("leave: exit pending\n");
		return;
	}

	if (!ipw_is_associated(priv)) {
		IPW_DEBUG_RATE("leave: not associated\n");
		return;
	}

	if (priv->iw_mode != IEEE80211_IF_TYPE_IBSS) {

		sta = sta_info_get(local, priv->bssid);
		if (sta && sta->rate_ctrl_priv) {
			scale_priv = (struct ipw_rate_scale_priv*)
				sta->rate_ctrl_priv;
			ctl = (struct ipw_rate_scale_data *)
			    &(scale_priv->win[0]);
			rc = ipw_rate_scale_flush_old(priv, ctl);
			sta_info_put(sta);
		} else if (sta)
			sta_info_put(sta);
	} else {
		struct ipw_station_entry *stations;
		int count = 0;

		stations = &(priv->stations[0]);
		for (i = STA_ID; i <
		     (priv->hw_setting.number_of_stations - 1); i++) {
			if ((stations[i].used)) {
				count++;
				sta = sta_info_get(local,
						   stations[i].sta.sta.MACAddr);

				if (sta && sta->rate_ctrl_priv) {
					scale_priv = (struct ipw_rate_scale_priv*)
						sta->rate_ctrl_priv;
					
					ctl = (struct ipw_rate_scale_data *)
						&(scale_priv->win[0]);
					rc |=
						ipw_rate_scale_flush_old(priv, ctl);

					sta_info_put(sta);
				} else if (sta)
					sta_info_put(sta);

			}
			if (count >= (priv->num_stations - 1))
				break;
		}
	}
	spin_lock_irqsave(&priv->lq_mngr.lock, flags);

	tx_ave = (priv->tx_packets - priv->lq_mngr.tx_packets)
	    + 1;

	priv->lq_mngr.tx_packets = priv->tx_packets + 1;

	if (rc && !(priv->status & STATUS_EXIT_PENDING)) {

		duration = priv->lq_mngr.flush_time * 1000 / HZ;

		IPW_DEBUG_RATE("Packets: %d duration: %d msec\n",
			       tx_ave, duration);
		if (duration == 0)
			tx_ave = 0;
		else
			tx_ave = (tx_ave * 1000) / duration;

		if (tx_ave == 0)
			duration = IPW_RATE_FLUSH_MAX;
		else {
			duration = (IPW_RATE_SCALE_AVG_PACKETS) / tx_ave;

			if (duration > IPW_RATE_FLUSH_MAX)
				duration = IPW_RATE_FLUSH_MAX;
			else if (duration < IPW_RATE_FLUSH_MIN)
				duration = IPW_RATE_FLUSH_MIN;
		}
		priv->lq_mngr.flush_time = duration * HZ / 1000;

		IPW_DEBUG_RATE("new flush period: %d msec ave %d\n",
			       duration, tx_ave);

		mod_timer(&priv->rate_scale_flush, jiffies +
			  priv->lq_mngr.flush_time);

		priv->lq_mngr.stamp = jiffies;
	}

	priv->lq_mngr.stamp_last = jiffies;

	spin_unlock_irqrestore(&priv->lq_mngr.lock, flags);
	IPW_DEBUG_RATE("leave\n");
}


static int ipw_rate_scale_init_handle(struct ipw_priv *priv, 
				      s32 window_size)
{
	int rc = 0;
	unsigned long flags;

	IPW_DEBUG_RATE("initialize rate scale window to %d\n", window_size);

	if ((window_size > IPW_RATE_SCALE_MAX_WINDOW) || (window_size < 0))
		window_size = IPW_RATE_SCALE_MAX_WINDOW;

	spin_lock_irqsave(&priv->lq_mngr.lock, flags);

	priv->lq_mngr.expected_tpt = NULL;
	priv->lq_mngr.next_higher_rate = NULL;
	priv->lq_mngr.next_lower_rate = NULL;
	priv->lq_mngr.stamp = jiffies;
	priv->lq_mngr.stamp_last = jiffies;

	priv->lq_mngr.flush_time = IPW_RATE_SCALE_FLUSH;
	priv->lq_mngr.tx_packets = 0;
	priv->lq_mngr.max_window_size = window_size;

	spin_unlock_irqrestore(&priv->lq_mngr.lock, flags);

	return rc;
}

static void ipw_collect_tx_data(struct ipw_priv *priv,
				struct ipw_rate_scale_data *window,
				int success, int retries)
{
	unsigned long flags;
	u64 mask;

	if ((priv->lq_mngr.max_window_size < 0) ||
	    (priv->lq_mngr.max_window_size > IPW_RATE_SCALE_MAX_WINDOW)) {
		IPW_DEBUG_RATE("leave: max_window_size index\n");
		return;
	}

	if (!retries) {
		IPW_DEBUG_RATE("leave: retries == 0 -- should be at "
			       "least 1\n");
		return;
	}

	while (retries--) {
		spin_lock_irqsave(&priv->lq_mngr.lock, flags);

		if (window->counter >= priv->lq_mngr.max_window_size) {
			window->counter = priv->lq_mngr.max_window_size;
			window->counter--;
			mask = (1 << (priv->lq_mngr.max_window_size - 1));
			if (window->data & mask) {
				window->data &= ~mask;
				window->success_counter--;
			}
		}

		window->counter++;
		mask = window->data;
		window->data = (mask << 1);
		if (success) {
			window->success_counter++;
			window->data |= 0x1;
		}

		if (window->counter > 0) {
			window->success_ratio = 128 * 
				(100 * window->success_counter) / 
				window->counter;
		} else
			window->success_ratio = IPW_INVALID_VALUE;

		window->stamp = jiffies;

		spin_unlock_irqrestore(&priv->lq_mngr.lock, flags);
	}
}

static int ipw_rate_scale_rxon_handle(struct ipw_priv *priv, s32 sta_id)
{
	int rc = 0;
	int i;
	s32 rssi = 0;
	struct ipw_tpt_entry *entry = NULL;
	unsigned long flags;
	struct ieee80211_conf *conf = NULL;
	struct ieee80211_local *local = wdev_priv(priv->net_dev->ieee80211_ptr);
	struct sta_info *sta;

	IPW_DEBUG_RATE("enter\n");

	conf = ieee80211_get_hw_conf(priv->ieee);

	if (!ipw_is_associated(priv)) {
		IPW_DEBUG_RATE("leave: not associated\n");
		return 0;
	}

	spin_lock_irqsave(&priv->lq_mngr.lock, flags);

	/* All but G protected use this rate fallback table */
	priv->lq_mngr.next_lower_rate = ipw_lower_rate_g;
	priv->lq_mngr.next_higher_rate = ipw_higher_rate_g;

	switch (priv->phymode) {
	case MODE_IEEE80211G:
		if (priv->active_rxon.flags & RXON_FLG_TGG_PROTECT_MSK) {
			priv->lq_mngr.expected_tpt = ipw_expected_tpt_g_prot;
			priv->lq_mngr.next_higher_rate = 
				ipw_higher_rate_g_prot;
			priv->lq_mngr.next_lower_rate = ipw_lower_rate_g_prot;
		} else 
			priv->lq_mngr.expected_tpt = ipw_expected_tpt_g;
		break;

	case MODE_IEEE80211B:
		priv->lq_mngr.expected_tpt = ipw_expected_tpt_b;
		break;

	case MODE_IEEE80211A:
		priv->lq_mngr.expected_tpt = ipw_expected_tpt_a;
		break;
	default:
		printk(KERN_WARNING "Invalid mode in rxon_handle\n");
		spin_unlock_irqrestore(&priv->lq_mngr.lock, flags);
		return -EINVAL;
	}
	
	rssi = priv->last_rx_rssi;
	if (rssi == 0)
		rssi = IPW_MIN_RSSI_VAL;
	
	IPW_DEBUG(IPW_DL_INFO | IPW_DL_RATE, "Network RSSI: %d\n", rssi);

	entry = ipw_get_tpt_by_rssi(rssi, priv->phymode);
	if (entry) 
		i = entry->rate_scale_index;
	else if (priv->phymode == MODE_IEEE80211A)
		i = RATE_SCALE_6M_INDEX;
	else
		i = RATE_SCALE_1M_INDEX;
	
	
	ipw_sync_station(priv, sta_id, ipw_rates[i].plcp,
			 CMD_ASYNC | CMD_NO_LOCK);

	sta = sta_info_get(local, priv->stations[sta_id].sta.sta.MACAddr);
	if (sta && sta->rate_ctrl_priv) {
		struct ipw_rate_scale_priv *scale_priv = 
			(void *)sta->rate_ctrl_priv;

		scale_priv->start_rate = i;

		sta_info_put(sta);
	}

	IPW_DEBUG_RATE("leave: rssi %d assign rate index: "
		       "%d (plcp 0x%x)\n",
		       rssi, i, ipw_rates[i].plcp);
	
	spin_unlock_irqrestore(&priv->lq_mngr.lock, flags);
	
	return rc;
}

static void rs_rate_init(void *priv_rate, void *priv_sta,
			 struct ieee80211_local *local,
			 struct sta_info *sta)
{
	int i;
	struct ipw_priv *priv = (struct ipw_priv *) priv_rate;

	IPW_DEBUG_RATE("enter\n");

	ipw_rate_scale_init_handle(priv, IPW_RATE_SCALE_MAX_WINDOW);

	/* TODO: what is a good starting rate for STA? About middle? Maybe not
	 * the lowest or the highest rate.. Could consider using RSSI from
	 * previous packets? Need to have IEEE 802.1X auth succeed immediately
	 * after assoc.. */

	if (priv && (priv->iw_mode == IEEE80211_IF_TYPE_IBSS)) {
		u8 sta_id;

		sta_id = ipw_add_station(priv, sta->addr,
					 0, (CMD_ASYNC | CMD_NO_LOCK));
	}

	for (i = IPW_MAX_RATES - 1; i >= 0; i--) {
		if ((sta->supp_rates & (1<<i)) &&
		    (priv->ieee_rates[i].flags & IEEE80211_RATE_SUPPORTED)) {
			sta->txrate = i;
			break;
		}
	}

	sta->last_txrate = sta->txrate;

#ifdef CONFIG_IWLWIFI_HT
	memcpy(&priv->current_assoc_ht, &sta->ht_info,
				sizeof(struct sta_ht_info));

	IPW_WARNING("XXXY is HT network %d \n", sta->ht_info.is_ht);
#endif

	IPW_DEBUG_RATE("leave\n");
}

static void *rs_alloc(struct ieee80211_local *local)
{
	IPW_DEBUG_RATE("enter\n");
	IPW_DEBUG_RATE("leave\n");
	return local->hw.priv;
}

static void rs_free(void *priv)
{
	IPW_DEBUG_RATE("NOP\n");
}

static void *rs_alloc_sta(void *priv, gfp_t gfp)
{
	struct ipw_rate_scale_priv *rctrl;
	int i;

	IPW_DEBUG_RATE("enter\n");

	rctrl = kzalloc(sizeof(struct ipw_rate_scale_priv),
			gfp);
	if (!rctrl) {
		IPW_DEBUG_RATE("leave: ENOMEM\n");
		return NULL;
	}

	rctrl->start_rate = IPW_INVALID_RATE;

	for (i = 0; i < IPW_MAX_RATES; i++)
		ipw_clear_window(&rctrl->win[i]);

	IPW_DEBUG_RATE("leave\n");

	return rctrl;
}

static void rs_free_sta(void *priv, void *priv_sta)
{
	struct ipw_rate_scale_priv *rctrl = priv_sta;
	
	IPW_DEBUG_RATE("enter\n");
	kfree(rctrl);
	IPW_DEBUG_RATE("leave\n");
}

static void rs_clear(void *priv)
{
	IPW_DEBUG_RATE("NOP\n");
}

static u8 ipw_get_next_rate(int phymode, u8 rate_index)
{
	if (rate_index == IPW_FIRST_OFDM_RATE) {
		/* If not 802.11a then return highest CCK rate */
		if (phymode != MODE_IEEE80211A)
			return IPW_LAST_CCK_RATE;
		/* else return first OFDM (lowest) rate */
		return IPW_FIRST_OFDM_RATE;
	}

	if (rate_index == IPW_FIRST_CCK_RATE)
		return IPW_FIRST_CCK_RATE;

	return rate_index - 1;
}

static void rs_tx_status(void *priv_rate,
			 struct net_device *dev,
			 struct sk_buff *skb,
			 struct ieee80211_tx_status *tx_resp)
{
	u8 retries, current_count;
	int scale_rate_index, first_index, last_index;
	unsigned long flags;
	struct sta_info *sta;
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)skb->data;
	struct ipw_priv *priv = (struct ipw_priv *)priv_rate;
	struct ieee80211_local *local = wdev_priv(dev->ieee80211_ptr);
	struct ipw_rate_scale_priv *scale_priv;

	IPW_DEBUG_RATE("enter\n");

	if (!ipw_is_associated(priv)) {
		IPW_DEBUG_RATE("leave: not associated\n");
		return;
	}

	retries = tx_resp->retry_count;

	first_index = tx_resp->control.tx_rate;
	if ((first_index < 0) || (first_index >= IPW_MAX_RATES)) {
		IPW_DEBUG_RATE("leave: Rate out of bounds: %0x for %d\n", 
			       tx_resp->control.tx_rate, first_index);
		return;
	}

	sta = sta_info_get(local, hdr->addr1);
	if (!sta || !sta->rate_ctrl_priv) {
		if (sta)
			sta_info_put(sta);
		IPW_DEBUG_RATE("leave: No STA priv data to update!\n");
		return;
	}
	
	scale_priv = (void *)sta->rate_ctrl_priv;

	scale_rate_index = first_index;
	last_index = first_index;

	/**
	 * Update the window for each rate (we determine which rates
	 * were Tx'd based on the total number of retries vs. the number
	 * of retries configured for each rate -- currently set to the
	 * priv value 'retry_rate' vs. rate specific 
	 * 
	 * On exit from this while loop last_index indicates the rate
	 * at which the frame was finally transmitted (or failed if no
	 * ACK)
	 */
	while (retries > 0) {
		if (retries < priv->retry_rate) {
			current_count = retries;
			last_index = scale_rate_index;
		} else {
			current_count = priv->retry_rate;
			last_index = ipw_get_next_rate(
				priv->phymode, scale_rate_index);
		}

		/* Update this rate accounting for as many retries
		 * as was used for it (per current_count) */
		ipw_collect_tx_data(priv, 
				    &scale_priv->win[scale_rate_index], 
				    0,
				    current_count);
		IPW_DEBUG_RATE("Update rate %d for %d retries.\n",
			       scale_rate_index, current_count);

		retries -= current_count;

		if (retries)
			scale_rate_index = ipw_get_next_rate(
				priv->phymode, scale_rate_index);
	}

	/* Update the last index window with success/failure based on ACK */
	IPW_DEBUG_RATE("Update rate %d with %s.\n",
		       last_index, 
		       (tx_resp->flags & IEEE80211_TX_STATUS_ACK) ?
		       "success" : "failure");
	ipw_collect_tx_data(priv, 
			    &scale_priv->win[last_index], 
			    tx_resp->flags & IEEE80211_TX_STATUS_ACK, 
			    1);

	sta_info_put(sta);

	spin_lock_irqsave(&priv->lq_mngr.lock, flags);

	/* Reschedule the rate scale flush again */
	if (time_after(jiffies, priv->lq_mngr.stamp +
		       priv->lq_mngr.flush_time)) {
		mod_timer(&priv->rate_scale_flush,
			  jiffies + IPW_RATE_SCALE_WIN_FLUSH);
		priv->lq_mngr.stamp = jiffies;

	}

	spin_unlock_irqrestore(&priv->lq_mngr.lock, flags);
	
	IPW_DEBUG_RATE("leave\n");

	return;
}



static struct ieee80211_rate *ipw_get_lowest_rate(struct ieee80211_local
						  *local)
{
	struct ieee80211_hw_mode *mode = local->oper_hw_mode;
	int i; 

	for (i = 0; i < mode->num_rates; i++) {
		struct ieee80211_rate *rate = &mode->rates[i];

		if (rate->flags & IEEE80211_RATE_SUPPORTED)
			return rate;
	}

	return &mode->rates[0];
}

static u16 ipw_get_adjacent_rate(struct ipw_priv *priv,
				 u8 index,
				 u16 rate_mask)
{
	u8 i;
	u8 high = IPW_INVALID_RATE;
	u8 low = IPW_INVALID_RATE;

	if (unlikely(priv->phymode == MODE_IEEE80211A)) {
		for (i = index; i > 0; i--) {
			if (rate_mask & (1 << (i - 1))) {
				low = i - 1;
				break;
			}
		}

		for (i = index + 1; i < IPW_MAX_RATES; i++) {
			if (rate_mask & (1 << i)) {
				high = i;
				break;
			}
		}

		return (high << 8) | low;
	}

	if (likely(priv->lq_mngr.next_lower_rate)) {
		low = index;
		while (low != IPW_INVALID_RATE) {
			low = priv->lq_mngr.next_lower_rate[low];
			if (low == IPW_INVALID_RATE)
				break;
			if (rate_mask & (1 << low)) 
				break;

			IPW_DEBUG_RATE("Skipping non supported low "
				       "rate %d\n", high);
		}
	}

	if (likely(priv->lq_mngr.next_higher_rate)) {
		high = index;
		while (high != IPW_INVALID_RATE) {
			high = priv->lq_mngr.next_higher_rate[high];
			if (high == IPW_INVALID_RATE)
				break;
			if (rate_mask & (1 << high)) 
				break;
			
			IPW_DEBUG_RATE("Skipping non supported high "
				       "rate %d\n", high);
		}
	}

	return (high << 8) | low;
}

/*
  find the highest available rate for REPLY_TX command.
*/
static struct ieee80211_rate *rs_get_rate(void *priv_rate,
					  struct net_device *dev,
					  struct sk_buff *skb,
					  struct rate_control_extra *extra)
{
	u8 low = IPW_INVALID_RATE;
	u8 high = IPW_INVALID_RATE;
	u16 high_low;
	int index;
	struct ipw_rate_scale_priv *scale_priv;
	struct ipw_rate_scale_data *window = NULL;
	int current_tpt = IPW_INVALID_VALUE;
	int low_tpt = IPW_INVALID_VALUE;
	int high_tpt = IPW_INVALID_VALUE;
	u32 fail_count;
	s8 scale_action = 0;
	unsigned long flags;
	struct ieee80211_local *local = wdev_priv(dev->ieee80211_ptr);
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)skb->data;
	struct sta_info *sta;
	u16 fc, rate_mask;
	struct ipw_priv *priv = (struct ipw_priv *)priv_rate;

	IPW_DEBUG_RATE("enter\n");

	memset(extra, 0, sizeof(*extra));

	fc = le16_to_cpu(hdr->frame_control);
	if ((WLAN_FC_GET_TYPE(fc) != IEEE80211_FTYPE_DATA) ||
	    (is_multicast_ether_addr(hdr->addr1))) {
		/* Send management frames and broadcast/multicast data using
		 * lowest rate. */
		/* TODO: this could probably be improved.. */
  		IPW_DEBUG_RATE("leave: lowest rate (not data or is "
			       "multicast)\n");
		
		return ipw_get_lowest_rate(local);
	}
	    
	sta = sta_info_get(local, hdr->addr1);
	if (!sta || !sta->rate_ctrl_priv) {
		IPW_DEBUG_RATE("leave: No STA priv data to update!\n");
		if (sta)
			sta_info_put(sta);
		return NULL;
	}

	rate_mask = sta->supp_rates;
	index = min(sta->txrate & 0xffff, IPW_MAX_RATES - 1);
	
	spin_lock_irqsave(&priv->lq_mngr.lock, flags);

	scale_priv = (struct ipw_rate_scale_priv*)
		sta->rate_ctrl_priv;

	if (scale_priv->start_rate != IPW_INVALID_RATE) {
		index = scale_priv->start_rate;
		scale_priv->start_rate = IPW_INVALID_RATE;
	}

	window = &(scale_priv->win[index]);

	fail_count = window->counter - window->success_counter;

	if (((fail_count <= IPW_RATE_SCALE_MIN_FAILURE_TH) &&
	     (window->success_counter < IPW_RATE_SCALE_MIN_SUCCESS_TH))
	    || (priv->lq_mngr.expected_tpt == NULL)) {

		window->average_tpt = IPW_INVALID_VALUE;
		spin_unlock_irqrestore(&priv->lq_mngr.lock, flags);

  		IPW_DEBUG_RATE("Invalid average_tpt on rate %d: "
			       "counter: %d, success_counter: %d, "
			       "expected_tpt is %sNULL\n", 
			       index,
			       window->counter, 
			       window->success_counter,
			       priv->lq_mngr.expected_tpt ? "not " : "");
		goto out;

	} 

	window->average_tpt = ((window->success_ratio *
				priv->lq_mngr.
				expected_tpt[index] + 64) / 128);

	high_low = ipw_get_adjacent_rate(priv, index, rate_mask);
	low = high_low & 0xff;
	high = (high_low >> 8) & 0xff;

	current_tpt = window->average_tpt;

	if (low  != IPW_INVALID_RATE) 
		low_tpt = scale_priv->win[low].average_tpt;

	if (high != IPW_INVALID_RATE)
		high_tpt = scale_priv->win[high].average_tpt;

	spin_unlock_irqrestore(&priv->lq_mngr.lock, flags);

	scale_action = 1;

	if ((window->success_ratio < IPW_RATE_SCALE_DECREASE_TH) ||
	    !current_tpt) {
		IPW_DEBUG_RATE("decrease rate because of low success_ratio\n");
		scale_action = -1;
	} else if ((low_tpt == IPW_INVALID_VALUE) &&
		   (high_tpt == IPW_INVALID_VALUE)) {
		scale_action = 1;
	} else if ((low_tpt != IPW_INVALID_VALUE) &&
		   (high_tpt != IPW_INVALID_VALUE)
		   && (low_tpt < current_tpt)
		   && (high_tpt < current_tpt)) {
		IPW_DEBUG_RATE("No action -- low [%d] & high [%d] < "
			       "current_tpt [%d]\n",
			       low_tpt, high_tpt, current_tpt);
		scale_action = 0;
	} else {
		if (high_tpt != IPW_INVALID_VALUE) {
			if (high_tpt > current_tpt)
				scale_action = 1;
			else {
				IPW_DEBUG_RATE
					("decrease rate because of high tpt\n");
				scale_action = -1;
			}
		} else if (low_tpt != IPW_INVALID_VALUE) {
			if (low_tpt > current_tpt) {
				IPW_DEBUG_RATE
					("decrease rate because of low tpt\n");
				scale_action = -1;
			} else
				scale_action = 1;
		}
	}

	if ((window->success_ratio > IPW_RATE_SCALE_HIGH_TH) ||
	    (current_tpt > (128 * window->average_tpt))) {
		IPW_DEBUG_RATE("No action -- success_ratio [%d] > HIGH_TH or "
			       "current_tpt [%d] > average_tpt [%d]\n",
			       window->success_ratio,
			       current_tpt, 128 * window->average_tpt);
		scale_action = 0;
	}

	switch (scale_action) {
	case -1:
		if (low != IPW_INVALID_RATE)
			index = low;
		break;

	case 1:
		if (high != IPW_INVALID_RATE)
			index = high;

		break;

	case 0:
	default:
		break;
	}

	IPW_DEBUG_RATE("Selected %d (action %d) - low %d high %d\n",
		       index, scale_action, low, high);

 out:
	    
	sta->last_txrate = index;
	sta->txrate = sta->last_txrate;
	sta_info_put(sta);

	IPW_DEBUG_RATE("leave: %d\n", index);
	return &priv->ieee_rates[index];
}


static char* rs_name = "iwlwifi rate-scale";
static void ipw_set_rate_scale_handlers(struct ipw_priv *priv)
{
	priv->rate_control.module = NULL;
	priv->rate_control.name = rs_name;
	priv->rate_control.tx_status = &rs_tx_status;
	priv->rate_control.get_rate = &rs_get_rate;
	priv->rate_control.rate_init = rs_rate_init;
	priv->rate_control.clear = rs_clear;
	priv->rate_control.alloc = rs_alloc;
	priv->rate_control.free = rs_free;
	priv->rate_control.alloc_sta = rs_alloc_sta;
	priv->rate_control.free_sta = rs_free_sta;
}
