#ifndef __iwl_3945_rs_h__
#define __iwl_3945_rs_h__

struct ipw_rate_info {
	u8 plcp;
	u8 ieee;
};

#define IPW_FIRST_OFDM_RATE 0
#define IPW_LAST_OFDM_RATE  7
#define IPW_FIRST_CCK_RATE  8
#define IPW_LAST_CCK_RATE   11

#define IPW_OFDM_RATE_6MB_MASK    (1<<0)
#define IPW_OFDM_RATE_9MB_MASK    (1<<1)
#define IPW_OFDM_RATE_12MB_MASK   (1<<2)
#define IPW_OFDM_RATE_18MB_MASK   (1<<3)
#define IPW_OFDM_RATE_24MB_MASK   (1<<4)
#define IPW_OFDM_RATE_36MB_MASK   (1<<5)
#define IPW_OFDM_RATE_48MB_MASK   (1<<6)
#define IPW_OFDM_RATE_54MB_MASK   (1<<7)
#define IPW_CCK_RATE_1MB_MASK     (1<<8)
#define IPW_CCK_RATE_2MB_MASK     (1<<9)
#define IPW_CCK_RATE_5MB_MASK     (1<<10)
#define IPW_CCK_RATE_11MB_MASK    (1<<11)

#define IPW_CCK_BASIC_RATES_MASK    \
       (IPW_CCK_RATE_1MB_MASK     | \
        IPW_CCK_RATE_2MB_MASK)

#define IPW_CCK_RATES_MASK          \
       (IPW_CCK_BASIC_RATES_MASK  | \
        IPW_CCK_RATE_5MB_MASK     | \
        IPW_CCK_RATE_11MB_MASK)

#define IPW_OFDM_BASIC_RATES_MASK   \
        (IPW_OFDM_RATE_6MB_MASK   | \
        IPW_OFDM_RATE_12MB_MASK   | \
        IPW_OFDM_RATE_24MB_MASK)

#define IPW_OFDM_RATES_MASK         \
       (IPW_OFDM_BASIC_RATES_MASK | \
        IPW_OFDM_RATE_9MB_MASK    | \
        IPW_OFDM_RATE_18MB_MASK   | \
        IPW_OFDM_RATE_36MB_MASK   | \
        IPW_OFDM_RATE_48MB_MASK   | \
        IPW_OFDM_RATE_54MB_MASK)

#define IPW_BASIC_RATES_MASK \
        (IPW_OFDM_BASIC_RATES_MASK | \
        IPW_CCK_BASIC_RATES_MASK)

#define IPW_DEFAULT_RATES_MASK ((1<<12)-1)


/* NOTE:
 *
 * Do not reorder the ipw_rates list --
 *
 * OFDM rates must come before CCK rates in order for routines like
 * reg_txpower_set_from_eeprom to be compatible with the uCode
 */
static struct ipw_rate_info ipw_rates[IPW_MAX_RATES] = {
	[0] = {13, 12},   /*  6mbps */
	[1] = {15, 18},   /*  9mbps */
	[2] = {5, 24},    /* 12mbps */
	[3] = {7, 36},    /* 18mbps */
	[4] = {9, 48},    /* 24mbps */
	[5] = {11, 72},   /* 36mbps */
	[6] = {1, 96},    /* 48mbps */
	[7] = {3, 108},    /* 54mbps */
	[8] = {10, 2},    /*  1mbps */
	[9] = {20, 4},    /*  2mbps */
	[10] = {55, 11},  /*5.5mbps */
	[11] = {110, 22}, /* 11mbps */
};

static inline int ipw_rate_index_from_plcp(int plcp)
{
	int i = 0;
	for (i = 0; i < ARRAY_SIZE(ipw_rates); i++) 
		if (ipw_rates[i].plcp == plcp)
			return i;
	return -1;
}

static int ipw_rate_scale_rxon_handle(struct ipw_priv *priv, s32 sta_id);

#endif
